/** SoA container for vectors and complex objects base on Vincenzo Innocente's ContainerSoa
 *
 */

#ifndef DATA_INCLUDE_CONTAINER_SOA_H_
#define DATA_INCLUDE_CONTAINER_SOA_H_

#include <vector>
#include <tuple>
#include <utility>
#include <type_traits>

#include "align_allocator.h"
#include "soa_iterator.h"
#include "utils/cpu_architecture.h"
#include "utils/metaprog.h"

#include "data/vector.h"

#include <assert.h>

namespace opice {
namespace data {
namespace container_soa {

namespace details {
    constexpr int align = opice::utils::ALIGNMENT_BYTES;
    template<typename T> using AlignedVector = std::vector<T,align_allocator<T, align>>;
}

template<typename T, int DIM, bool = std::is_arithmetic<T>::value>
class ContainerSoa {
};

// Check if the type T has a member SOA_VDIM. If it does, we presume it's a SoA-ready vector
template<typename T, int DIM, bool = std::is_integral<decltype(T::SOA_VDIM)>::value>
struct ContainerSoaTraits {};

// A general type has to define the SOATUPLE
template<typename T, int DIM>
struct ContainerSoaTraits<T, DIM, false> {
    using VAL   = typename T::VAL;
    using CREF  = typename T::CREF;
    using REF   = typename T::REF;
    using SOATUPLE = typename T::SOATUPLE;
};

// A vector has an auto-generated SOATUPLE deduced from its dimension and inner value type
template<typename T, int DIM>
struct ContainerSoaTraits<T, DIM, true> {
    using VAL   = typename T::VAL;
    using CREF  = typename T::CREF;
    using REF   = typename T::REF;
    using InnerContainer = ContainerSoa<typename T::value,DIM>;
    using SOATUPLE = typename utils::gen_tuple<T::SOA_VDIM, InnerContainer>::type;
};


/*** Multidimensional container for compound types
 *
 *   @tparam T      enclosed type
 *   @tparam DIM    number of dimensions of the container
 */
template<typename T, int DIM>
class ContainerSoa<T, DIM, false> {
public:
    using VAL  = typename ContainerSoaTraits<T,DIM>::VAL;
    using CREF = typename ContainerSoaTraits<T,DIM>::CREF;
    using REF  = typename ContainerSoaTraits<T,DIM>::REF;
    using Data = typename ContainerSoaTraits<T,DIM>::SOATUPLE;

    using value_type = VAL;
    using iterator       = SoaIterator<REF,  ContainerSoa<T, DIM, false>>;
    using const_iterator = SoaIterator<CREF, ContainerSoa<T, DIM, false>>;

    template<typename Tuple, std::size_t ... I>
    void resize_impl(Tuple dims, std::index_sequence<I...>) {
        using swallow = int[];
        // The void cast causes unpacked .resize_from_tuple() functions to be called.
        static_cast<void>(swallow { 0, (
                static_cast<void>(std::get<I>(_data).resize_from_tuple(dims)) , 0
                )... });
    }

    template<typename ... Args>
    void resize(Args ... n) {
        _n = utils::multiply_arg_pack_int(n...); // size of the flat array - all dimensions multiplied
        auto dimensions = std::make_tuple(n...);
        resize_impl( dimensions,
                std::make_integer_sequence<std::size_t,
                        std::tuple_size<Data>::value> { });
    }

    ContainerSoa() {
    }

    template<typename ... Args>
    explicit ContainerSoa(int n0, Args ... n) { //TODO: This should be a pack of integers, not arbitrary types
        resize(n0, n...);
    }

    /*** Given a vector and an integer sequence, resizes the multidimensional container
     *   to the size specified by the vector.
     *    */
    template <typename V, size_t... Is>
    void resize_by_vector(V v, std::index_sequence<Is...> ) {
        resize(v[Is]...);   // Will call resize(v[0],v[1],v[2],...). For v=(5,6,7): resize(5,6,7)
    }

    template<typename VT, bool VSOA>
    ContainerSoa(const vector::Vector<VT, DIM, VSOA>& v){
        resize_by_vector(v, std::make_index_sequence<DIM>{} );
    }

    auto size() const {
        return _n;
    }

    iterator begin() {
        return iterator(*this);
    }
    iterator end() {
        return iterator(*this, size());
    }

    const_iterator begin() const {
        return const_iterator(*this);
    }
    const_iterator end() const {
        return const_iterator(*this, size());
    }


    // Indexing by vector
    template<typename V, std::size_t ... I, typename IDX>
    V tuple_to_vector(IDX idx, std::index_sequence<I...>) {
        return V(std::get<I>(_data)[idx] ...);
    }

    // Indexing by a single integer - calls operator[]
    template<typename V, std::size_t ... I>
    V tuple_to_vector(unsigned int j, std::index_sequence<I...>) {
        return V(std::get<I>(_data)[j] ...);
    }

    REF operator[](unsigned int i) {
        assert(i < _n);
        return tuple_to_vector<REF>(i,
                std::make_integer_sequence<std::size_t,
                        std::tuple_size<Data>::value> { });
    }

    CREF operator[](unsigned int i) const {
        assert(i < _n);
        return tuple_to_vector<CREF>(i,
                std::make_integer_sequence<std::size_t,
                        std::tuple_size<Data>::value> { });
    }



    /*** Access by index vector - uses operator() ***/

    template<typename VT, bool VSOA>
    REF operator[](const vector::Vector<VT, DIM, VSOA>& v) {
        return tuple_to_vector<REF>(v,
                std::make_integer_sequence<std::size_t,
                        std::tuple_size<Data>::value> { });
    }

    template<typename VT, bool VSOA>
    CREF operator[] (const vector::Vector<VT, DIM, VSOA>& v) const {
        return tuple_to_vector<CREF>(v,
                std::make_integer_sequence<std::size_t,
                        std::tuple_size<Data>::value> { });
    }




    Data _data;
    unsigned int _n = 0;
    //std::array<int, DIM> _size;     // Container size, accessed at runtime
};

/*** Multidimensional container for arithmetic types
 *
 *   @tparam T      enclosed type
 *   @tparam DIM    number of dimensions of the container
 */
template<typename T, int DIM>
class ContainerSoa<T, DIM, true> {
public:

    using value = typename std::remove_const_t<typename std::remove_reference_t<typename std::remove_const_t<T>>>;
    using ref  = typename std::add_lvalue_reference_t<T>;
    using cref = typename std::add_lvalue_reference_t<typename std::add_const_t<T>>;
    using VAL  = value;
    using CREF = cref;
    using REF  = ref;

    using value_type = VAL;

    using Flat = details::AlignedVector<VAL>;

    //TODO: The flat array must have at least DIM elements, otherwise multi constructor fails
    //TODO: Document why exactly do we need the no-parameter constructor
    ContainerSoa() :
            _flat(DIM) {
    }

    //TODO: Write a test for this
    /*** Constructs a multidimensional array, with dimension lengths specified as separate arguments
     *
     *   This method initializes the underlying ContainerSoa::Flat storage of a size computed by
     *   multiplying all constructor arguments. the ContainerSoa::Multi DIM-dimensional view is then
     *   initialized from the flat storage
     *
     *   @tparam Args variadic argument pack
     *   @param  n    variadic - lengths in succesive dimensions
     */
    template<typename ... Args> ContainerSoa(Args ... n) { //TODO: This should be a pack of integers, not arbitrary types
        resize(n...);
    }

    REF operator[](std::size_t i) {
        assert(i < _flat.size());
        return _flat[i];
    }

    CREF operator[](std::size_t i) const {
        assert(i < _flat.size());
        return _flat[i];
    }

    template <typename V> //TODO: Check performance
    REF index_by_vector(const V &index) {
        std::size_t idx = 0;
        for (int i = 0; i<DIM; ++i){    //TODO: Replace with a lambda?
            idx += index[i]*std::accumulate(_size.begin()+i+1, _size.end(), 1, std::multiplies<std::size_t>());
        }
        return _flat[idx];
    }

    //TODO: Rip it out
    template<typename VT, bool VSOA>
    REF operator[](const vector::Vector<VT, DIM, VSOA>& index) {
        std::size_t idx = 0;
        for (int i = 0; i<DIM; ++i){    //TODO: Replace with a lambda?
            assert(index[i] >= 0);
            assert(static_cast<std::size_t>(index[i]) < _size[i]);
            idx += index[i]*std::accumulate(_size.begin()+i+1, _size.end(), 1, std::multiplies<std::size_t>());
        }
        return _flat[idx];
    }

    template<typename VT, bool VSOA>
    CREF operator[](const vector::Vector<VT, DIM, VSOA>& index) const {
        std::size_t idx = 0;
        for (int i = 0; i<DIM; ++i){    //TODO: Replace with a lambda?
            assert(index[i] >= 0);
            assert(static_cast<std::size_t>(index[i]) < _size[i]);
            idx += index[i]*std::accumulate(_size.begin()+i+1, _size.end(), 1, std::multiplies<std::size_t>());
        }
        return _flat[idx];
    }

    template<typename ... Args> void resize(Args ... n) {
        _size = {static_cast<std::size_t>(n)...};
        const int flatsize = utils::multiply_arg_pack_int(n...); // size of the _flat array - all dimensions multiplied
        _flat.resize(flatsize);
    }

    template<typename Tuple> void resize_from_tuple(Tuple&& t){
        constexpr auto iseq = std::make_integer_sequence<std::size_t, std::tuple_size<typename std::decay<Tuple>::type>::value> { };
        resize_tuple_caller( std::forward<Tuple>(t), iseq );
    }

    template<typename Tuple, std::size_t... I>
    void resize_tuple_caller(Tuple&& t, std::index_sequence<I...>)
    {
      resize(std::get<I>(std::forward<Tuple>(t))...);
    }

private:
    Flat _flat;
    std::array<std::size_t, DIM> _size;     // Container size, accessed at runtime
};

}}} // namespace opice::data::container_soa


#include "data/container_free_ops.h"


#endif /* DATA_INCLUDE_CONTAINER_SOA_H_ */
