/*
 * vector_impl_3d.h
 *
 *  Created on: Apr 7, 2016
 *      Author: strazce
 */

#ifndef INCLUDE_DATA_VECTOR_IMPL_3D_H_
#define INCLUDE_DATA_VECTOR_IMPL_3D_H_

#include <cmath>
#include <algorithm>
#include <type_traits>
#include <cstddef>
#include <ostream>

#include "vector_details.h"
#include "vector_expressions.h"

namespace opice {
namespace data {
namespace vector {

template<typename T>
struct alignas(ALIGN) Vector<T, 3, false>
    : VectorBase<Vector<T, 3, false>, T>,
      Expr<Vector<T, 3, false>, T, 3>
{
    static constexpr unsigned int DIM = 3;

    using value = typename Details<T>::value;
    using ref = typename Details<T>::ref;
    using cref = typename Details<T>::cref;
    using VAL = Vector<value, 3, false>;
    using CREF = Vector<cref, 3, false>;
    using REF = Vector<ref, 3, false>;

    T x, y, z;

    constexpr Vector() = delete;
    constexpr Vector(T ix, T iy, T iz) :
            x(ix), y(iy), z(iz) {
    }
    template<typename T2, bool SOA_READY>
    constexpr Vector(Vector<T2, DIM, SOA_READY> v) :
            x(v.x), y(v.y), z(v.z) {
    }
    template<typename T2, bool SOA_READY>
    Vector& operator=(Vector<T2, DIM, SOA_READY> && v) {
        x = v.x;
        y = v.y;
        z = v.z;
        return *this;
    }
    constexpr Vector(Vector const & v) :
            x(v.x), y(v.y), z(v.z) {
    }

    /// @internal Pointer arithmetics validity must be checked compile-time
    cref ret_op(unsigned int k) const {
        return *(&x + k);
    }

    ref ret_op(unsigned int k) {
        return *(&x + k);
    }

    template<typename T2, bool SOA2>
    constexpr bool operator==(const Vector<T2,3,SOA2>& b){
        return x == b.x && y == b.y && z == b.z;
    }

    static constexpr inline auto Zeros() { return Vector(0,0,0); }
    static constexpr inline auto Ones()  { return Vector(1,1,1); }

};

template<typename T>
struct alignas(ALIGN) Vector<T, 3, true>
    : VectorBase<Vector<T, 3, true>, T>,
      Expr<Vector<T, 3, true>, T, 3>
{
    static constexpr unsigned int DIM = 3;
    static constexpr unsigned int SOA_VDIM = 3;

    using value = typename Details<T>::value;
    using ref = typename Details<T>::ref;
    using cref = typename Details<T>::cref;
    using VAL = Vector<value, 3, true>;
    using CREF = Vector<cref, 3, true>;
    using REF = Vector<ref, 3, true>;

    T x, y, z;

    constexpr Vector() = delete;

    constexpr Vector(T ix, T iy, T iz) :
            x(ix), y(iy), z(iz) {
    }

    template<typename T2, bool SOA_READY>
    constexpr Vector(Vector<T2, DIM, SOA_READY> v) :
            x(v.x), y(v.y), z(v.z) {
    }

    constexpr Vector(Vector const & v) :
            x(v.x), y(v.y), z(v.z) {
    }

    template<typename T2, bool SOA_READY>
    Vector& operator=(Vector<T2, DIM, SOA_READY> v) {
        x = v.x;
        y = v.y;
        z = v.z;
        return *this;
    }

    template<typename T2, bool SOA_READY>
    Vector& operator=(const Vector<T2&, DIM, SOA_READY> & v) {
        x = v.x;
        y = v.y;
        z = v.z;
        return *this;
    }

    cref ret_op(unsigned int k) const {
        return (k == 0) ? x : (k == 1 ? y : z);
    }

    ref ret_op(unsigned int k) {
        return (k == 0) ? x : (k == 1 ? y : z);
    }

    template<typename T2, bool SOA2>
    constexpr bool operator==(const Vector<T2,3,SOA2>& b){
        return x == b.x && y == b.y && z == b.z;
    }


    static constexpr inline auto Zeros() { return Vector(0,0,0); }
    static constexpr inline auto Ones()  { return Vector(1,1,1); }

};

/// @internal Compile-time checks for pointer arithmetics validity
//TODO: Generalize the assertions, or move them to some more sensible place
using NonSoaV3f = Vector<float, 3, false>;
using NonSoaV3d = Vector<double, 3, false>;
static_assert(offsetof(NonSoaV3f,y) ==   sizeof(float), "Offset of y is not at float size");
static_assert(offsetof(NonSoaV3f,z) == 2*sizeof(float), "Offset of z is not at 2*float size");
static_assert(offsetof(NonSoaV3d,y) ==   sizeof(double), "Offset of y is not at double size");
static_assert(offsetof(NonSoaV3d,z) == 2*sizeof(double), "Offset of z is not at 2*double size");

}}} // namespace opice::data::vector



#endif /* INCLUDE_DATA_VECTOR_IMPL_3D_H_ */
