/*
 * vector_impl_3d.h
 *
 *  Created on: Apr 7, 2016
 *      Author: strazce
 */

#ifndef INCLUDE_DATA_VECTOR_IMPL_4D_H_
#define INCLUDE_DATA_VECTOR_IMPL_4D_H_

#include <cmath>
#include <algorithm>
#include <type_traits>
#include <cstddef>
#include <ostream>

#include "vector_details.h"
#include "vector_expressions.h"

namespace opice {
namespace data {
namespace vector {

template<typename T>
struct alignas(ALIGN) Vector<T, 4, false>
    : VectorBase<Vector<T, 4, false>, T>,
      Expr<Vector<T, 4, false>, T, 4>
{
    static constexpr unsigned int DIM = 4;

    using value = typename Details<T>::value;
    using ref = typename Details<T>::ref;
    using cref = typename Details<T>::cref;
    using VAL = Vector<value, DIM, false>;
    using CREF = Vector<cref, DIM, false>;
    using REF = Vector<ref, DIM, false>;

    T x0, x1, x2, x3;

    constexpr Vector() = delete;
    constexpr Vector(T ix0, T ix1, T ix2, T ix3) :
            x0(ix0), x1(ix1), x2(ix2), x3(ix3) {
    }
    template<typename T2, bool SOA_READY>
    constexpr Vector(Vector<T2, DIM, SOA_READY> v) :
            x0(v.x0), x1(v.x1), x2(v.x2), x3(v.x3) {
    }
    template<typename T2, bool SOA_READY>
    Vector& operator=(Vector<T2, DIM, SOA_READY> && v) {
        x0 = v.x0;
        x1 = v.x1;
        x2 = v.x2;
        x3 = v.x3;
        return *this;
    }
    constexpr Vector(Vector const & v) :
        x0(v.x0), x1(v.x1), x2(v.x2), x3(v.x3) {
    }

    /// @internal Pointer arithmetics validity must be checked compile-time
    cref ret_op(unsigned int k) const {
        return *(&x0 + k);
    }

    ref ret_op(unsigned int k) {
        return *(&x0 + k);
    }

    template<typename T2, bool SOA2>
    constexpr bool operator==(const Vector<T2,4,SOA2>& b){
        return x0 == b.x0 && x1 == b.x1 && x2 == b.x2 && x3 == b.x3;
    }

    static constexpr inline auto Zeros() { return Vector(0,0,0,0); }
    static constexpr inline auto Ones()  { return Vector(1,1,1,1); }

};

template<typename T>
struct alignas(ALIGN) Vector<T, 4, true>
    : VectorBase<Vector<T, 4, true>, T>,
      Expr<Vector<T, 4, true>, T, 4>
{
    static constexpr unsigned int DIM = 4;
    static constexpr unsigned int SOA_VDIM = 4;

    using value = typename Details<T>::value;
    using ref = typename Details<T>::ref;
    using cref = typename Details<T>::cref;
    using VAL = Vector<value, DIM, true>;
    using CREF = Vector<cref, DIM, true>;
    using REF = Vector<ref, DIM, true>;

    T x0, x1, x2, x3;

    constexpr Vector() = delete;
    constexpr Vector(T ix0, T ix1, T ix2, T ix3) :
            x0(ix0), x1(ix1), x2(ix2), x3(ix3) {
    }
    template<typename T2, bool SOA_READY>
    constexpr Vector(Vector<T2, DIM, SOA_READY> v) :
            x0(v.x0), x1(v.x1), x2(v.x2), x3(v.x3) {
    }
    template<typename T2, bool SOA_READY>
    Vector& operator=(Vector<T2, DIM, SOA_READY> && v) {
        x0 = v.x0;
        x1 = v.x1;
        x2 = v.x2;
        x3 = v.x3;
        return *this;
    }
    constexpr Vector(Vector const & v) :
        x0(v.x0), x1(v.x1), x2(v.x2), x3(v.x3) {
    }

    cref ret_op(unsigned int k) const {
        return k==0 ? x0 : (k==1 ? x1 : (k==2 ? x2 : x3));
    }

    ref ret_op(unsigned int k) {
        return k==0 ? x0 : (k==1 ? x1 : (k==2 ? x2 : x3));
    }

    template<typename T2, bool SOA2>
    constexpr bool operator==(const Vector<T2,4,SOA2>& b){
        return x0 == b.x0 && x1 == b.x1 && x2 == b.x2 && x3 == b.x3;
    }


    static constexpr inline auto Zeros() { return Vector(0,0,0,0); }
    static constexpr inline auto Ones()  { return Vector(1,1,1,1); }

};

/// @internal Compile-time checks for pointer arithmetics validity
//TODO: Generalize the assertions, or move them to some more sensible place
using NonSoaV4f = Vector<float,  4, false>;
using NonSoaV4d = Vector<double, 4, false>;
static_assert(offsetof(NonSoaV4f,x1) ==   sizeof(float), "Offset of x1 is not at float size");
static_assert(offsetof(NonSoaV4f,x2) == 2*sizeof(float), "Offset of x2 is not at 2*float size");
static_assert(offsetof(NonSoaV4f,x3) == 3*sizeof(float), "Offset of x3 is not at 2*float size");
static_assert(offsetof(NonSoaV4d,x1) ==   sizeof(double), "Offset of x1 is not at double size");
static_assert(offsetof(NonSoaV4d,x2) == 2*sizeof(double), "Offset of x2 is not at 2*double size");
static_assert(offsetof(NonSoaV4d,x3) == 3*sizeof(double), "Offset of x3 is not at 2*double size");

}}} // namespace opice::data::vector



#endif /* INCLUDE_DATA_VECTOR_IMPL_3D_H_ */
