/*
 * stepper.h
 *
 *  Created on: Jun 14, 2013
 *      Author: vyskocil
 */

#ifndef INCLUDE_PHYSICS_STEPPER_H_
#define INCLUDE_PHYSICS_STEPPER_H_

#include "common.h"
#include "utils/basic_logger.h"

namespace opice {
namespace physics {

/** @brief Interface for working with space-time step. Every class which needs access to dt, dx, dy, dz should inherit from this.
 *
 */
class Stepper {
public:
    Stepper() = delete;

    Stepper(  Scalar dt
            , Scalar dx
            , Scalar dy
            #if DIM_POS == 3
            , Scalar dz
            #endif
            )
     :    _dt(dt),
          _dt_factor(1),
          _dx(dx),
          _dy(dy)
       #if DIM_POS == 3
        , _dz(dz)
        , _dr(VPOS(dx,dy,dz))
        , _ds(dt, dx, dy, dz)
       #elif DIM_POS == 2
        , _dr(VPOS(dx,dy))
        , _ds(dt, dx, dy)
       #endif
        ,_idr(VPOS::Zeros()), _istep(VSTI::Zeros())
    { recalculate_inverses(); }

    Stepper(Scalar dt, VPOS &dr)
      :   _dt(dt), _dt_factor(1)
        , _dx(dr.x), _dy(_dr.y)
        #if DIM_POS == 3
        , _dz(dr.z)
        #endif
        , _dr(dr)
        #if DIM_POS == 3
        , _ds(dt, dr.x, dr.y, dr.z)
        #elif DIM_POS == 2
        , _ds(dt, dr.x, dr.y)
        #endif
        ,_idr(VPOS::Zeros()), _istep(VSTI::Zeros())
    { recalculate_inverses(); }

    Stepper(const VSTI& s)
      :   _dt(s[0]),
          _dt_factor(1),
          _dx(s[1]),
          _dy(s[2])
        #if DIM_POS == 3
        , _dz(s[3]),
          _dr(VPOS(s[1],s[2],s[3]))
        #elif DIM_POS == 2
        , _dr(VPOS(s[1],s[2]))
        #endif
        ,_ds(s)
        ,_idr(VPOS::Zeros()), _istep(VSTI::Zeros())
    { recalculate_inverses(); }

    void set_dt(Scalar dt){ _dt = dt; _ds[0] = dt; recalculate_inverses();};
    void set_dx(Scalar dx){ _dx = dx; _dr.x = dx; _ds[1] = dx; recalculate_inverses();};
    void set_dy(Scalar dy){ _dy = dy; _dr.y = dy; _ds[2] = dy; recalculate_inverses();};
    #if DIM_POS == 3
    void set_dz(Scalar dz){ _dz = dz; _dr.z = dz; _ds[3] = dz; recalculate_inverses();};
    #endif

    const VSTI& step(){ return _ds; }

    void reset_cfl(Scalar factor){
        _dt_factor = factor;
        recalculate_cfl();
    }

    void reset_cfl(){
        recalculate_cfl();
    }

    void print_step_info(){
        LI <<   "dt = " << _dt
           << "; dx = " << _dx
           << "; dy = " << _dy
           #if DIM_POS == 3
           << "; dz = " << _dz
           #endif
           ;
    }

protected:
    Scalar _dt;
    Scalar _dt_factor;

    Scalar _dx, _dy;
    #if DIM_POS == 3
    Scalar _dz;
    #endif

    VPOS _dr;
    VSTI _ds;

    Scalar _idt;
    Scalar _idx, _idy;
    #if DIM_POS == 3
    Scalar _idz;
    #endif

    VPOS _idr;
    VSTI _istep;

    void recalculate_inverses(){
        _idt = 1/_dt;

//        _idr = _dr.cwiseInverse(); //TODO: Implement cwiseInverse
        #if DIM_POS == 3
        _idr = VPOS(1/_dr.x, 1/_dr.y, 1/_dr.z);
        #elif DIM_POS == 2
        _idr = VPOS(1/_dr.x, 1/_dr.y);
        #endif

        _idx = _idr.x;
        _idy = _idr.y;
        #if DIM_POS == 3
        _idz = _idr.z;
        #endif
        _istep = VSTI(_idt
                    , _idx
                    , _idy
                    #if DIM_POS == 3
                    , _idz
                    #endif
                    );
    }

    void recalculate_cfl(){

        #if DIM_POS == 3
        set_dt(_dt_factor / sqrt( 1/sqr(_dx) + 1/sqr(_dy) + 1/sqr(_dz) ));
        #elif DIM_POS == 2
        set_dt(_dt_factor / sqrt( 1/sqr(_dx) + 1/sqr(_dy) ));
        #endif

    }
};

}
}

#endif /* STEPPER_H_ */
