/** Runs all tests - wrapper needed for Eclipse unit testing framework
 *  which is only able to run one C++ executable at a time
 *
 *  Created on: Mar 26, 2016
 *      Author: strazce
 */

#include <cstdlib>
#include <vector>
#include <string>
#include <algorithm>
#include <cstdio>
#include <memory>
#include <iostream>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>

using std::string;
using std::vector;
using std::accumulate;

namespace fs = boost::filesystem;

// List of filenames of the unit test
const vector<string> test_executables {
  "unit_v1",
  "unit_v2",
  "unit_v3",
  "unit_v4",
  "unit_c"
};

/* Run one test and strip the tags */
void exec_and_replace(const char* cmd) {
    std::shared_ptr<FILE> pipe(popen(cmd, "r"), pclose);
    if (!pipe) std::cout << "ERROR";

    const string tag_open = "<TestLog>";
    const string tag_close = "</TestLog>";

    char buffer[128];
    string current = "";
    string previous = "";
    string writeout = "";
    while (!feof(pipe.get())) {
        /* We have to test whether the output contains the opening and closing tags, and strip them.
         * Since the tags could possibly be split across two buffers, we always save the previous buffer,
         * and concatenate it with the current one. This way we're sure we catch the tag if it's there
         * (of course provided the buffer is long enough). If the tag wasn't there, we return the previous
         * buffer, and continue with our loop. If we find the tag, we remove it from the concatenated string,
         * return the whole thing and empty the buffer before next iteration.  */
        if (fgets(buffer, 128, pipe.get()) != NULL){
            previous = current;
            current  = buffer;
            string test = previous + current;
            if(test.find(tag_open) != string::npos){    //TODO: This only happens at the beginning, so we could find a better way to remove it, and not have to test it here.
                test.erase(test.find(tag_open), tag_open.length());
                std::cout << test;
                current = "";
            } else if(test.find(tag_close) != string::npos) {
                test.erase(test.find(tag_close), tag_close.length());
                std::cout << test;
                current = "";
            } else {
                std::cout << previous;
            }
        }
    }
    std::cout << current;
}

int main(int argc, char** argv){

    /* Eclipse test runner adds additional arguments to the boost tests,
     * we want to pass them to individual test */
    string args;
    if (argc > 1){
        vector<string> all_args(argv + 1, argv + argc); // Get arguments passed by Eclipse
        args = accumulate(all_args.begin(), all_args.end(), string(""),
                [](const string& a, const string& b){
                    return a + " " + b;
                }); // Concatenate them into single space-separated string
    } else {
        args = "";
    }

    /* We want to get the absolute path of the directory where the tests are located.
       Otherwise we would have to properly set the working directory in Eclipse */
    fs::path full_path( fs::initial_path<fs::path>() );
    full_path = fs::system_complete( fs::path( argv[0] ) );
    string directory = full_path.remove_filename().string();

    /* Each test outputs a separate XML file, but need to output just one valid XML file
     * for all of them. We add the root tag to the output stream, and strip the root
     * tags form each individual test. */
    std::cout << "<TestLog>";
    for(auto test: test_executables){
        string cmd = directory + "/" + test + " " + args;
        exec_and_replace(cmd.c_str());
    }
    std::cout << "</TestLog>" << std::flush;

    return 0;
}







