/*
 * vector3d.cpp
 *
 *  Created on: Feb 11, 2016
 *      Author: strazce
 */



#define BOOST_TEST_MODULE ContainerSoA
#include <boost/test/unit_test.hpp>

#include <boost/mpl/back_inserter.hpp>
#include <boost/mpl/insert_range.hpp>
#include <boost/mpl/list.hpp>

#include "../../src/include/data/vector.h"
#include "data/container_soa.h"

namespace mpl = boost::mpl;

template<typename T> using V3 = opice::data::vector::Vector<T,3,true>;
template<typename T, int DIM> using CONT3 = opice::data::container_soa::ContainerSoa<V3<T>, DIM>;

using V3f = V3<float>;

template<typename T> using V4 = opice::data::vector::Vector<T,4,true>;
template<typename T, int DIM> using CONT4 = opice::data::container_soa::ContainerSoa<V4<T>, DIM>;

using V4f = V4<float>;

using float_types = mpl::list<float, double, long double>;
using int_types = mpl::list<short, int, long>;

using arithmetic_types = mpl::insert_range<
                                float_types,
                                mpl::end<float_types>::type,
                                int_types>::type;

template<typename T1, typename T2>
T1 S(T2 t) {
    return static_cast<T1>(t);
}

constexpr int NUM = 1000;
constexpr int XMAX = 10;
constexpr int YMAX = 10;
constexpr int ZMAX = 10;

template < typename T, typename F >
void static VEC_EQ(V3<T> v, F x, F y, F z){
    BOOST_CHECK(v.x == x);
    BOOST_CHECK(v.y == y);
    BOOST_CHECK(v.z == z);
}

template<typename T, typename F>
void static VEC_EQ_ALL(V3<T>& v, F x){
    VEC_EQ(v, x, x, x);
}

template < typename T, typename F >
void static VEC_EQ(V4<T> v, F x0, F x1, F x2, F x3){
    BOOST_CHECK(v[0] == x0);
    BOOST_CHECK(v[1] == x1);
    BOOST_CHECK(v[2] == x2);
    BOOST_CHECK(v[3] == x3);
}

template<typename T, typename F>
void static VEC_EQ_ALL(V4<T>& v, F x){
    VEC_EQ(v, x, x, x, x);
}

BOOST_AUTO_TEST_SUITE(Construction)

BOOST_AUTO_TEST_CASE_TEMPLATE(Construct1D, T, arithmetic_types){
    CONT3<T,1> cont(NUM);
    cont[NUM-1] = V3<T>(1,2,3);
    VEC_EQ(cont[NUM-1],1,2,3);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(Construct2D, T, arithmetic_types){
    CONT3<T,2> cont(XMAX,YMAX);
    cont[(XMAX-1)*(YMAX-1)] = V3<T>(1,2,3);
    VEC_EQ(cont[(XMAX-1)*(YMAX-1)],1,2,3);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(Construct3D, T, arithmetic_types){
    CONT3<T,3> cont(XMAX,YMAX,ZMAX);
    cont[(XMAX-1)*(YMAX-1)*(ZMAX-1)] = V3<T>(1,2,3);
    VEC_EQ(cont[(XMAX-1)*(YMAX-1)*(ZMAX-1)],1,2,3);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(Construct3D_V4, T, arithmetic_types){
    CONT4<T,3> cont(XMAX,YMAX,ZMAX);
    cont[(XMAX-1)*(YMAX-1)*(ZMAX-1)] = V4<T>(1,2,3,4);
    VEC_EQ(cont[(XMAX-1)*(YMAX-1)*(ZMAX-1)],1,2,3,4);
}


BOOST_AUTO_TEST_CASE_TEMPLATE(ElementAccess1D, T, arithmetic_types){

    CONT3<T,1> cont(NUM);

    for(auto x: cont){
        x.x=1; x.y=2; x.z=3;
    }
    BOOST_TEST_CHECKPOINT("Initialized " << NUM << " elements.");

    for(auto x: cont){
        VEC_EQ(x, S<T>(1), S<T>(2), S<T>(3));
    }

    for(std::size_t i=1; i<cont.size(); ++i){
        cont[i] += cont[i-1];
    }
    BOOST_TEST_CHECKPOINT("Performed addition.");

    for(std::size_t i=1; i<cont.size(); ++i){
        VEC_EQ(cont[i], S<T>(i+1), S<T>(2*(i+1)), S<T>(3*(i+1)) );
    }
}


BOOST_AUTO_TEST_CASE_TEMPLATE(ElementAccess2D, T, arithmetic_types){

    CONT3<T,2> cont(XMAX,YMAX);

    for(auto x: cont){
        x.x=1; x.y=2; x.z=3;
    }
    BOOST_TEST_CHECKPOINT("Initialized " << NUM << " elements.");

    for(auto x: cont){
        VEC_EQ(x, S<T>(1), S<T>(2), S<T>(3));
    }

    for(std::size_t i=1; i<cont.size(); ++i){
        cont[i] += cont[i-1];
    }
    BOOST_TEST_CHECKPOINT("Performed addition.");

    for(std::size_t i=1; i<cont.size(); ++i){
        VEC_EQ(cont[i], S<T>(i+1), S<T>(2*(i+1)), S<T>(3*(i+1)) );
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(ElementAccess3D, T, arithmetic_types){

    CONT3<T,3> cont(XMAX,YMAX,ZMAX);

    for(auto x: cont){
        x.x=1; x.y=2; x.z=3;
    }
    BOOST_TEST_CHECKPOINT("Initialised " << NUM << " elements.");

    for(auto x: cont){
        VEC_EQ(x, S<T>(1), S<T>(2), S<T>(3));
    }

    for(std::size_t i=1; i<cont.size(); ++i){
        cont[i] += cont[i-1];
    }
    BOOST_TEST_CHECKPOINT("Performed addition.");

    for(std::size_t i=1; i<cont.size(); ++i){
        VEC_EQ(cont[i], S<T>(i+1), S<T>(2*(i+1)), S<T>(3*(i+1)) );
    }
}

BOOST_AUTO_TEST_CASE_TEMPLATE(ElementAccess3D_V4, T, arithmetic_types){

    CONT4<T,3> cont(XMAX,YMAX,ZMAX);

    for(auto x: cont){
        x[0]=1; x[1]=2; x[2]=3; x[3]=4;
    }
    BOOST_TEST_CHECKPOINT("Initialised " << NUM << " elements.");

    for(auto x: cont){
        VEC_EQ(x, S<T>(1), S<T>(2), S<T>(3), S<T>(4));
    }

    for(std::size_t i=1; i<cont.size(); ++i){
        cont[i] += cont[i-1];
    }
    BOOST_TEST_CHECKPOINT("Performed addition.");

    for(std::size_t i=1; i<cont.size(); ++i){
        VEC_EQ(cont[i], S<T>(i+1), S<T>(2*(i+1)), S<T>(3*(i+1)), S<T>(4*(i+1)) );
    }
}

BOOST_AUTO_TEST_SUITE_END()
